export module pige.utility:scope_guard;

import std;

namespace pige::utility
{
	export template<class F>
	class scope_guard
	{
		F m_f;
		bool m_active;
	public:
		explicit scope_guard(F f) noexcept
			: m_f{ std::move(f) }, m_active(true)
		{ }
		scope_guard() = delete;
		scope_guard(const scope_guard &) = delete;

		scope_guard(scope_guard &&other) noexcept
			requires std::move_constructible<F>
			: m_f(std::move(other.m_f)), m_active(other.m_active)
		{
			other.m_active = false;
		}
		scope_guard &operator=(const scope_guard &) = delete;

		scope_guard &operator=(scope_guard &&other) noexcept
			requires std::swappable<F>
		{
			std::swap(m_f, other.m_f);
			std::swap(m_active, other.m_active);
			return *this;
		}

		~scope_guard() { if (m_active) m_f(); }
	};
}