module;

#include <windows.h>

export module pige.regions;

export import pige.winapi;

export namespace pige::regions
{
	class teardrop : public winapi::window
	{
	public:

		teardrop(HINSTANCE inst, LPCWSTR title)
			: window{
				inst,
				winapi::window_desc{
					title,
					{ CW_USEDEFAULT, CW_USEDEFAULT },
					{ CW_USEDEFAULT, CW_USEDEFAULT },
					WS_OVERLAPPEDWINDOW | WS_CLIPCHILDREN
			  } },
			child{
			  inst,
			  winapi::window_desc{
				  L"Child",
				  { 300, 200 },
				  { 250, 150 },
				  WS_OVERLAPPEDWINDOW | WS_CHILD | WS_VISIBLE,
				  0,
				  *this
			} },
			popup{
			  inst,
			  winapi::window_desc{
				  L"Popup",
				  { 350, 250 },
				  { CW_USEDEFAULT, CW_USEDEFAULT },
				  static_cast<LONG>(WS_POPUP | WS_OVERLAPPEDWINDOW | WS_VISIBLE),
				  0,
				  *this
			} }
		{}

	protected:
		LRESULT window_proc(UINT msg, WPARAM wparam, LPARAM lparam) override
		{
			switch (msg)
			{
			case WM_ERASEBKGND:
				return 0;
			case WM_PAINT:
				on_paint();
				return 0;
			case WM_WINDOWPOSCHANGED:
			{
				auto &pos = *reinterpret_cast<LPWINDOWPOS>(lparam);
				if ((pos.flags & SWP_SHOWWINDOW) || !(pos.flags & SWP_NOSIZE))
					set_window_rgn();
				break;
			}
			}
			return window::window_proc(msg, wparam, lparam);
		}

	private:

		window child, popup;

		void set_window_rgn()
		{
			auto dc = GetWindowDC(*this);
			BeginPath(dc);
			RECT rc{};
			GetWindowRect(*this, &rc);

			float width = static_cast<float>(rc.right - rc.left);
			float height = static_cast<float>(rc.bottom - rc.top);
			POINTFLOAT cptsf[10]{
				//{0,0},{0,0},
				{ width,  0 },{ width / 8.f, 0 },

				{ 0,  5.f * height / 16.f },
				{ 0, 10.f * height / 16.f },
				{ 0, 15.f * height / 16.f },
				{ 3.f * width / 16.f, height },
				{ 6.f * width / 16.f, height },
				{ 9.f * width / 16.f, height },
				{ width, 15.f * height / 16.f },
				{ width, 0 }
			};
			POINT cpts[10]{};
			for (size_t i = 0; i < 10; ++i)
				cpts[i] = { static_cast<int>(round(cptsf[i].x)), static_cast<int>(round(cptsf[i].y)) };
			PolyBezier(dc, cpts, 10);
			EndPath(dc);
			HRGN rgn = PathToRegion(dc);
			SetWindowRgn(*this, rgn, true);
			ReleaseDC(*this, dc);
		}

		void on_paint()
		{
			//BeginPaint validates the update area, remember call GetUpdateRect/GetUpdateRgn beforehand
			RECT updt{}, cl{};
			GetUpdateRect(*this, &updt, false);
			PAINTSTRUCT ps;
			auto dc = BeginPaint(*this, &ps);
			auto mdc = CreateCompatibleDC(dc);
			RECT mrc{ 0, 0, updt.right - updt.left, updt.bottom - updt.top };
			auto oldbmp = SelectObject(mdc, CreateCompatibleBitmap(dc, mrc.right, mrc.bottom));
			auto bgbr = CreateSolidBrush(RGB(100, 120, 230));
			FillRect(mdc, &mrc, bgbr);
			DeleteObject(bgbr);

			RECT rc{ 850, 150, 1150, 350 }, temp;
			if (IntersectRect(&temp, &rc, &updt))
			{
				OffsetRect(&rc, -updt.left, -updt.top);
				auto oldbr = SelectObject(mdc, CreateSolidBrush(RGB(200, 210, 250)));
				Ellipse(mdc, rc.left, rc.top, rc.right, rc.bottom);
				DeleteObject(SelectObject(mdc, oldbr));
			}
			BitBlt(dc, updt.left, updt.top, mrc.right, mrc.bottom, mdc, 0, 0, SRCCOPY);
			DeleteObject(SelectObject(mdc, oldbmp));
			DeleteDC(mdc);
			EndPaint(*this, &ps);
		}
	};
}